package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.experiment.setup.config.ProvidedUrlRequestRSpec;
import be.iminds.ilabt.jfed.experiment.setup.config.Slice;
import be.iminds.ilabt.jfed.ui.cli2.instruction.Instruction;
import be.iminds.ilabt.jfed.ui.cli2.instruction.RunInstruction;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import org.apache.commons.cli.*;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;

public class ExperimenterCliTest {
    private final static ObjectMapper MAPPER = new ObjectMapper();
    private final static ObjectMapper YMLMAPPER = new ObjectMapper(new YAMLFactory());

    @org.testng.annotations.Test
    public void testPrintAction1() throws Exception {
        File configFile = File.createTempFile("test-config-pa", ".cfg");
        IOUtils.stringToFile(configFile, "{\n" +
                "    \"action\": \"run\",\n" +
                "    \"experiment\": {\n" +
                "        \"slice\": {\n" +
                "            \"sliceName\": \"tst2\"\n" +
                "        },\n" +
                "        \"requestRSpec\": {\n" +
                "            \"providedContentSource\": \"https://flsmonitor.fed4fire.eu/dock1.rspec\",\n" +
                "            \"source\": \"PROVIDE_URL\"\n" +
                "        }\n" +
                "    },\n" +
                "    \"debugOutputs\": [{\n" +
                "        \"level\": \"DEBUG\",\n" +
                "        \"format\": \"TXT\",\n" +
                "        \"target\": \"STDERR\"\n" +
                "    }]\n" +
                "}");
        configFile.deleteOnExit();

        String[] args = {
                "--print-action",
                "-p",
                "dummy.pem",
                "--action",
                configFile.getAbsolutePath()
        };
        final ByteArrayOutputStream baOut = new ByteArrayOutputStream();
        final ByteArrayOutputStream baErr = new ByteArrayOutputStream();
        final PrintStream out = new PrintStream(baOut);
        final PrintStream err = new PrintStream(baErr);
        final InputStream in = new ByteArrayInputStream(new byte[] {});
        BaseCli baseCli = new BaseCli(err);

        int exitVal = ExperimenterCli.main_internal(args, out, err, in, baseCli);

        String outString = new String(baOut.toByteArray(), StandardCharsets.UTF_8);
        String errString = new String(baErr.toByteArray(), StandardCharsets.UTF_8);

        System.out.println("exitVal: "+exitVal);
        System.out.println("Stderr: "+errString);
        System.out.println("Stdout: "+outString);

        assertThat(exitVal, is(0));
        assertThat(errString, is(equalTo("")));
        assertThat(outString.length(), greaterThan(2));

        Instruction instruction = YMLMAPPER.readValue(outString, Instruction.class);
        assertThat(instruction.getAction(), is(Instruction.Action.RUN));
        assertThat(instruction, instanceOf(RunInstruction.class));
        RunInstruction ri = (RunInstruction) instruction;
        assertThat(ri.getExperiment().getSlice().getSliceName(), is(equalTo("tst2")));
        assertThat(ri.getExperiment().getSlice().getProjectSource(), is(Slice.ProjectSource.AUTO));
        assertThat(ri.getExperiment().getRequestRSpec(), instanceOf(ProvidedUrlRequestRSpec.class));
        assertThat(((ProvidedUrlRequestRSpec)ri.getExperiment().getRequestRSpec())
                .getProvidedContentSource(), is(equalTo("https://flsmonitor.fed4fire.eu/dock1.rspec")));
    }

    @org.testng.annotations.Test
    public void testPrintAction2() throws Exception {
        File configFile = File.createTempFile("test-config-pa", ".cfg");
        IOUtils.stringToFile(configFile, "{\n" +
                "    \"action\": \"run\",\n" +
                "    \"experiment\": {\n" +
                "        \"slice\": {\n" +
                "            \"sliceName\": \"tst2\"\n" +
                "        },\n" +
                "        \"requestRSpec\": {\n" +
                "            \"providedContentSource\": \"https://flsmonitor.fed4fire.eu/dock1.rspec\",\n" +
                "            \"source\": \"PROVIDE_URL\"\n" +
                "        }\n" +
                "    },\n" +
                "    \"debugOutputs\": [{\n" +
                "        \"level\": \"DEBUG\",\n" +
                "        \"format\": \"TXT\",\n" +
                "        \"target\": \"STDERR\"\n" +
                "    }]\n" +
                "}");
        configFile.deleteOnExit();

        String[] args = {
                "--print-action",
                "--slice",
                "tst1",
                "-p",
                "dummy.pem",
                "--action",
                configFile.getAbsolutePath()
        };
        final ByteArrayOutputStream baOut = new ByteArrayOutputStream();
        final ByteArrayOutputStream baErr = new ByteArrayOutputStream();
        final PrintStream out = new PrintStream(baOut);
        final PrintStream err = new PrintStream(baErr);
        final InputStream in = new ByteArrayInputStream(new byte[] {});
        BaseCli baseCli = new BaseCli(err);

        int exitVal = ExperimenterCli.main_internal(args, out, err, in, baseCli);

        String outString = new String(baOut.toByteArray(), StandardCharsets.UTF_8);
        String errString = new String(baErr.toByteArray(), StandardCharsets.UTF_8);

        System.out.println("exitVal: "+exitVal);
        System.out.println("Stderr: "+errString);
        System.out.println("Stdout: "+outString);

        assertThat(exitVal, is(0));
        assertThat(errString, is(equalTo("")));
        assertThat(outString.length(), greaterThan(2));

        Instruction instruction = YMLMAPPER.readValue(outString, Instruction.class);
        assertThat(instruction.getAction(), is(Instruction.Action.RUN));
        assertThat(instruction, instanceOf(RunInstruction.class));
        RunInstruction ri = (RunInstruction) instruction;
        assertThat(ri.getExperiment().getSlice().getSliceName(), is(equalTo("tst1")));
        assertThat(ri.getExperiment().getSlice().getProjectSource(), is(Slice.ProjectSource.AUTO));
        assertThat(ri.getExperiment().getRequestRSpec(), instanceOf(ProvidedUrlRequestRSpec.class));
        assertThat(((ProvidedUrlRequestRSpec)ri.getExperiment().getRequestRSpec())
                .getProvidedContentSource(), is(equalTo("https://flsmonitor.fed4fire.eu/dock1.rspec")));
    }

    @org.testng.annotations.Test
    public void testPrintAction3() throws Exception {
        File configFile = File.createTempFile("test-config-pa", ".cfg");
        IOUtils.stringToFile(configFile, "{\n" +
                "    \"action\": \"run\",\n" +
                "    \"experiment\": {\n" +
                "        \"slice\": {\n" +
                "            \"sliceName\": \"tst2\"\n" +
                "        },\n" +
                "        \"requestRSpec\": {\n" +
                "            \"providedContentSource\": \"https://flsmonitor.fed4fire.eu/dock1.rspec\",\n" +
                "            \"source\": \"PROVIDE_URL\"\n" +
                "        }\n" +
                "    },\n" +
                "    \"debugOutputs\": [{\n" +
                "        \"level\": \"DEBUG\",\n" +
                "        \"format\": \"TXT\",\n" +
                "        \"target\": \"STDERR\"\n" +
                "    }]\n" +
                "}");
        configFile.deleteOnExit();

        String[] args = {
                "-p",
                "dummy.pem",
                "--print-action",
                "--slice",
                "tst1",
                "--project-name",
                "dummyProject",
                "--expiration-hours",
                "5",
                "--action",
                configFile.getAbsolutePath()
        };
        final ByteArrayOutputStream baOut = new ByteArrayOutputStream();
        final ByteArrayOutputStream baErr = new ByteArrayOutputStream();
        final PrintStream out = new PrintStream(baOut);
        final PrintStream err = new PrintStream(baErr);
        final InputStream in = new ByteArrayInputStream(new byte[] {});
        BaseCli baseCli = new BaseCli(err);

        int exitVal = ExperimenterCli.main_internal(args, out, err, in, baseCli);

        String outString = new String(baOut.toByteArray(), StandardCharsets.UTF_8);
        String errString = new String(baErr.toByteArray(), StandardCharsets.UTF_8);

        System.out.println("exitVal: "+exitVal);
        System.out.println("Stderr: "+errString);
        System.out.println("Stdout: "+outString);

        assertThat(exitVal, is(0));
        assertThat(errString, is(equalTo("")));
        assertThat(outString.length(), greaterThan(2));

        Instruction instruction = YMLMAPPER.readValue(outString, Instruction.class);
        assertThat(instruction.getAction(), is(Instruction.Action.RUN));
        assertThat(instruction, instanceOf(RunInstruction.class));
        RunInstruction ri = (RunInstruction) instruction;
        assertThat(ri.getExperiment().getSlice().getSliceName(), is(equalTo("tst1")));
        assertThat(ri.getExperiment().getSlice().getProject(), is(equalTo("dummyProject")));
        assertThat(ri.getExperiment().getSlice().getProjectSource(), is(Slice.ProjectSource.PROVIDED));
        assertThat(ri.getExperiment().getSlice().getExpireTimeMin(), is(5*60L));
        assertThat(ri.getExperiment().getRequestRSpec(), instanceOf(ProvidedUrlRequestRSpec.class));
        assertThat(((ProvidedUrlRequestRSpec)ri.getExperiment().getRequestRSpec())
                .getProvidedContentSource(), is(equalTo("https://flsmonitor.fed4fire.eu/dock1.rspec")));
    }

    @org.testng.annotations.Test
    public void testPrintAction4() throws Exception {
        String[] args = {
                "-p",
                "dummy.pem",
                "--action",
                "run",
                "--print-action",
                "--slice",
                "tst1",
                "--project-name",
                "dummyProject",
                "--expiration-hours",
                "5",
                "--rspec",
                "https://flsmonitor.fed4fire.eu/dock1.rspec"
        };
        final ByteArrayOutputStream baOut = new ByteArrayOutputStream();
        final ByteArrayOutputStream baErr = new ByteArrayOutputStream();
        final PrintStream out = new PrintStream(baOut);
        final PrintStream err = new PrintStream(baErr);
        final InputStream in = new ByteArrayInputStream(new byte[] {});
        BaseCli baseCli = new BaseCli(err);

        int exitVal = ExperimenterCli.main_internal(args, out, err, in, baseCli);

        String outString = new String(baOut.toByteArray(), StandardCharsets.UTF_8);
        String errString = new String(baErr.toByteArray(), StandardCharsets.UTF_8);

        System.out.println("exitVal: "+exitVal);
        System.out.println("Stderr: "+errString);
        System.out.println("Stdout: "+outString);

        assertThat(exitVal, is(0));
        assertThat(errString, is(equalTo("")));
        assertThat(outString.length(), greaterThan(2));

        Instruction instruction = YMLMAPPER.readValue(outString, Instruction.class);
        assertThat(instruction.getAction(), is(Instruction.Action.RUN));
        assertThat(instruction, instanceOf(RunInstruction.class));
        RunInstruction ri = (RunInstruction) instruction;
        assertThat(ri.getExperiment().getSlice().getSliceName(), is(equalTo("tst1")));
        assertThat(ri.getExperiment().getSlice().getProject(), is(equalTo("dummyProject")));
        assertThat(ri.getExperiment().getSlice().getProjectSource(), is(Slice.ProjectSource.PROVIDED));
        assertThat(ri.getExperiment().getSlice().getExpireTimeMin(), is(5*60L));
        assertThat(ri.getExperiment().getRequestRSpec(), instanceOf(ProvidedUrlRequestRSpec.class));
        assertThat(((ProvidedUrlRequestRSpec)ri.getExperiment().getRequestRSpec())
                .getProvidedContentSource(), is(equalTo("https://flsmonitor.fed4fire.eu/dock1.rspec")));
    }


    @org.testng.annotations.Test
    public void testCliArg() throws Exception {
        Options options = new Options();
        options.addOption(OptionBuilder.withLongOpt("test")
                .withDescription("Test")
                .hasArg()
                .withArgName("TEST")
                .create());

        CommandLineParser parser = new BasicParser();
        String[] args = new String[2];
        args[0] = "--test";
        args[1] = "SUCCESS";
        CommandLine line = parser.parse(options, args);

        assertThat(line.hasOption("test"), is(true));
        assertThat(line.getOptionValue("test"), is(equalTo(args[1])));
    }

    @org.testng.annotations.Test
    public void testCliMultiArg() throws Exception {
        Options options = new Options();
        options.addOption(OptionBuilder.withLongOpt("test")
                .withDescription("Test")
                .hasArg()
                .withArgName("TEST")
                .create());

        CommandLineParser parser = new BasicParser();
        String[] args = new String[4];
        args[0] = "--test";
        args[1] = "SUCCESS";
        args[2] = "--test";
        args[3] = "SUCCESS2";
        CommandLine line = parser.parse(options, args);

        assertThat(line.hasOption("test"), is(true));
        String [] actual = line.getOptionValues("test");
        assertThat(actual.length, is(equalTo(2)));
        assertThat(actual[0], is(equalTo(args[1])));
        assertThat(actual[1], is(equalTo(args[3])));
    }

    @org.testng.annotations.Test
    public void testCliArgs() throws Exception {
        Options options = new Options();
        options.addOption(OptionBuilder.withLongOpt("test")
                .withDescription("Test")
                .hasArgs(2)
                .create());

        CommandLineParser parser = new BasicParser();
        String[] args = new String[3];
        args[0] = "--test";
        args[1] = "ONE";
        args[2] = "TWO";
        CommandLine line = parser.parse(options, args);

        assertThat(line.hasOption("test"), is(true));
        System.out.println("test=" + Arrays.asList(line.getOptionValues("test")));
        assertThat(line.getOptionValue("test"), is(equalTo(args[1])));
    }

    //hasArgs(2) doesn't work like I would want  (--test <A> <B> is the same as --test <A> --test <B> and it should not be.)

//    @org.testng.annotations.Test(expectedExceptions = { ParseException.class })
//    public void testCliArgsBad() throws Exception {
//        Options options = new Options();
//        options.addOption(OptionBuilder.withLongOpt("test")
//                .withDescription("Test")
//                .hasArgs(2)
//                .create());
//
//        CommandLineParser parser = new BasicParser();
//        String[] args = new String[2];
//        args[0] = "--test";
//        args[1] = "ONE";
//        CommandLine line = parser.parse(options, args);
//
//        assertThat(line.hasOption("test"), is(true));
//        System.out.println("test=" + Arrays.asList(line.getOptionValues("test")));
//    }
//
//    @org.testng.annotations.Test(expectedExceptions = { ParseException.class })
//    public void testCliArgsBad2() throws Exception {
//        Options options = new Options();
//        options.addOption(OptionBuilder.withLongOpt("test")
//                .withDescription("Test")
//                .hasArgs(2)
//                .create());
//
//        CommandLineParser parser = new BasicParser();
//        String[] args = new String[4];
//        args[0] = "--test";
//        args[1] = "ONE";
//        args[2] = "TWO";
//        args[3] = "BAD";
//        CommandLine line = parser.parse(options, args);
//
//        assertThat(line.hasOption("test"), is(true));
//        System.out.println("test=" + Arrays.asList(line.getOptionValues("test")));
//    }
//
//    @org.testng.annotations.Test(expectedExceptions = { ParseException.class })
//    public void testCliArgsBad3() throws Exception {
//        Options options = new Options();
//        options.addOption(OptionBuilder.withLongOpt("test")
//                .withDescription("Test")
//                .hasArgs(2)
//                .create());
//
//        CommandLineParser parser = new BasicParser();
//        String[] args = new String[6];
//        args[0] = "--test";
//        args[1] = "ONE";
//        args[2] = "TWO";
//        args[3] = "BAD";
//        args[4] = "--test";
//        args[5] = "FOUR";
//        CommandLine line = parser.parse(options, args);
//
//        assertThat(line.hasOption("test"), is(true));
//        System.out.println("test=" + Arrays.asList(line.getOptionValues("test")));
//    }

    @org.testng.annotations.Test
    public void testCliMultiArgs() throws Exception {
        Options options = new Options();
        options.addOption(OptionBuilder.withLongOpt("test")
                .withDescription("Test")
                .hasArgs(2)
                .create());

        CommandLineParser parser = new BasicParser();
        String[] args = new String[6];
        args[0] = "--test";
        args[1] = "ONE";
        args[2] = "TWO";
        args[3] = "--test";
        args[4] = "THREE";
        args[5] = "FOUR";
        CommandLine line = parser.parse(options, args);

        assertThat(line.hasOption("test"), is(true));
        System.out.println("test=" + Arrays.asList(line.getOptionValues("test")));
        assertThat(line.getOptionValue("test"), is(equalTo(args[1])));
    }
}
